unit myhashedstringlist;

interface

uses
  Classes, IniFiles, SysUtils;

type
{copied from IniFiles in D7}
  { TStringHash - used internally by TMemIniFile to optimize searches. }

  PPHashItem = ^PHashItem;
  PHashItem = ^THashItem;
  THashItem = record
    Next: PHashItem;
    Key: string;
    Value: Integer;
  end;

  TStringHash = class
  private
    Buckets: array of PHashItem;
  protected
    function Find(const Key: string): PPHashItem;
    function HashOf(const Key: string): Cardinal; virtual;
  public
    constructor Create(Size: Cardinal = 256);
    destructor Destroy; override;
    procedure Add(const Key: string; Value: Integer);
    procedure Clear;
    procedure Remove(const Key: string);
    function Modify(const Key: string; Value: Integer): Boolean;
    function ValueOf(const Key: string): Integer;
  end;


  // TODO: support delete and other cases,
  // either via a total refresh or a more
  // precise operation on the hash table
  TMyHashedStringList = class (TStringList)
  private
    FValueHash: TStringHash;
    FNameHash: TStringHash;
    FSorted: Boolean;
    procedure SetSorted(const Value: Boolean);
    procedure UpdateValueHash;
    procedure UpdateNameHash;
    procedure UpdateSingleHash(S: string; nPos: Integer);
  public
    function Add(const S: string): Integer; override;
    destructor Destroy; override;
    function IndexOf(const S: string): Integer; override;
    function IndexOfName(const Name: string): Integer; override;
    property Sorted: Boolean read FSorted write SetSorted;
    function AddObject(const S: String; AObject: TObject): Integer;
      override;
  end;

implementation

{ TMyHashedStringList }

function TMyHashedStringList.Add(const S: string): Integer;
begin
  Result := inherited Add (s);
  if fSorted then
    UpdateSingleHash (S, Result);
end;

procedure TMyHashedStringList.UpdateSingleHash (S: string; nPos: Integer);
var
  P: Integer;
  Key: string;
begin
  // full value
  if not CaseSensitive then
    FValueHash.Add(AnsiUpperCase(S), nPos)
  else
    FValueHash.Add(S, nPos);
  // name
  P := AnsiPos('=', S);
  if P <> 0 then
  begin
    if not CaseSensitive then
      Key := AnsiUpperCase(Copy(S, 1, P - 1))
    else
      Key := Copy(S, 1, P - 1);
    FNameHash.Add(Key, nPos);
  end;
end;

function TMyHashedStringList.AddObject(const S: String;
  AObject: TObject): Integer;
begin
  Result := inherited AddObject (s, AObject);
  if fSorted then
    UpdateSingleHash (S, Result);
end;

destructor TMyHashedStringList.Destroy;
begin
  FValueHash.Free;
  FNameHash.Free;
  inherited Destroy;
end;

function TMyHashedStringList.IndexOf(const S: string): Integer;
begin
  if fSorted then
  begin
    if not CaseSensitive then
      Result :=  FValueHash.ValueOf(AnsiUpperCase(S))
    else
      Result :=  FValueHash.ValueOf(S);
  end
  else
    Result := inherited IndexOf (S);
end;

function TMyHashedStringList.IndexOfName(const Name: string): Integer;
begin
  if fSorted then
  begin
    if not CaseSensitive then
      Result := FNameHash.ValueOf(AnsiUpperCase(Name))
    else
      Result := FNameHash.ValueOf(Name);
  end
  else
    Result := inherited IndexOfName (Name);
end;

procedure TMyHashedStringList.SetSorted(const Value: Boolean);
begin
  FSorted := Value;
  if FSorted then
  begin
    // optimize: separate sorting flag
    UpdateNameHash;
    UpdateValueHash;
  end;
end;

procedure TMyHashedStringList.UpdateNameHash;
var
  I: Integer;
  P: Integer;
  Key: string;
begin
  if FNameHash = nil then
    FNameHash := TStringHash.Create (2048)
  else
    FNameHash.Clear;
  for I := 0 to Count - 1 do
  begin
    Key := Get(I);
    P := AnsiPos('=', Key);
    if P <> 0 then
    begin
      if not CaseSensitive then
        Key := AnsiUpperCase(Copy(Key, 1, P - 1))
      else
        Key := Copy(Key, 1, P - 1);
      FNameHash.Add(Key, I);
    end;
  end;
end;

procedure TMyHashedStringList.UpdateValueHash;
var
  I: Integer;
begin
  if FValueHash = nil then
    FValueHash := TStringHash.Create (2048)
  else
    FValueHash.Clear;
  for I := 0 to Count - 1 do
    if not CaseSensitive then
      FValueHash.Add(AnsiUpperCase(Self[I]), I)
    else
      FValueHash.Add(Self[I], I);
end;


{ TStringHash -> From D7 }

procedure TStringHash.Add(const Key: string; Value: Integer);
var
  Hash: Integer;
  Bucket: PHashItem;
begin
  Hash := HashOf(Key) mod Cardinal(Length(Buckets));
  New(Bucket);
  Bucket^.Key := Key;
  Bucket^.Value := Value;
  Bucket^.Next := Buckets[Hash];
  Buckets[Hash] := Bucket;
end;

procedure TStringHash.Clear;
var
  I: Integer;
  P, N: PHashItem;
begin
  for I := 0 to Length(Buckets) - 1 do
  begin
    P := Buckets[I];
    while P <> nil do
    begin
      N := P^.Next;
      Dispose(P);
      P := N;
    end;
    Buckets[I] := nil;
  end;
end;

constructor TStringHash.Create(Size: Cardinal);
begin
  inherited Create;
  SetLength(Buckets, Size);
end;

destructor TStringHash.Destroy;
begin
  Clear;
  inherited Destroy;
end;

function TStringHash.Find(const Key: string): PPHashItem;
var
  Hash: Integer;
begin
  Hash := HashOf(Key) mod Cardinal(Length(Buckets));
  Result := @Buckets[Hash];
  while Result^ <> nil do
  begin
    if Result^.Key = Key then
      Exit
    else
      Result := @Result^.Next;
  end;
end;

function TStringHash.HashOf(const Key: string): Cardinal;
var
  I: Integer;
begin
  Result := 0;
  for I := 1 to Length(Key) do
    Result := ((Result shl 2) or (Result shr (SizeOf(Result) * 8 - 2))) xor
      Ord(Key[I]);
end;

function TStringHash.Modify(const Key: string; Value: Integer): Boolean;
var
  P: PHashItem;
begin
  P := Find(Key)^;
  if P <> nil then
  begin
    Result := True;
    P^.Value := Value;
  end
  else
    Result := False;
end;

procedure TStringHash.Remove(const Key: string);
var
  P: PHashItem;
  Prev: PPHashItem;
begin
  Prev := Find(Key);
  P := Prev^;
  if P <> nil then
  begin
    Prev^ := P^.Next;
    Dispose(P);
  end;
end;

function TStringHash.ValueOf(const Key: string): Integer;
var
  P: PHashItem;
begin
  P := Find(Key)^;
  if P <> nil then
    Result := P^.Value
  else
    Result := -1;
end;


end.
